_ENV = AIContext

local function is_combat_ship(ship)
  for w in weapons(ship)
    if alive_weapons(w) > 0
      return true
    end
  end
  return false
end

-- don't count ships whose weapons are destroyed, as combat ships
function enemy_combat_ships()
  local f = current_enemy_ships ()
  return function(s,v)
    local a = f(s,v)
    while a
      if is_combat_ship(a) return a end
      --print('skipping',a)
      a=f(s,a)
    end
  end
end

-- TODO factor in weapon mods if needed
-- average damage per mount
function average_weapon_damage(weapon)
  if weapon.is_fighter
    --while hangars bays are not counted for attack value, handling hangars here
    --is important for deciding whether or not to reserve interceptors
    local ship = weapon.object
    local fighter_weapon = get_fighter_weapon(weapon.parent_type, ship.empire, ship)
    return WEAPONS[fighter_weapon].damage_fun.average
  end

  return weapon.info.damage_fun.average
end

-- average damage per subobject
function average_tracker_damage(tracker)
  --print("TRACKER", tracker.name, tracker.is_missile, tracker.is_fighter)
  if(tracker.is_missile)
    if(tracker.submunitions > 0)
      return WEAPONS[tracker.submunitions_weapon].damage_fun.average * tracker.submunitions
    end
    return WEAPONS[tracker.warhead.parent_type].damage_fun.average
  end
  
  if(tracker.is_fighter)
    local total = 0
    --fighters only ever have 1 weapon equipped at a time right now, but loop through weapons just in case...
    for _,wname in pairs(tracker.weapons) do
      total += WEAPONS[wname.parent_type].damage_fun.average
    end
    return total
  end

  return 0
end

local function alive_and_present(w)
  local alive = alive_weapons(w)
  if w.is_fighter
    return max(0, alive - w.missing)
  end
  return alive
end

-- a score that represents how "dangerous" a potential target is
function calc_attack_value(ship)
  local score = 0

  for w in weapons(ship) do
    -- NOTE the damage value of fightercraft are only counted when they are in their hangars
    -- it does mean that carriers drop in priority once their craft launch, but that's preferable to
    -- having carriers who have no more craft left retaining priority as if they did
    local alive = alive_and_present(w)
    local damage = average_weapon_damage(w)

    if w.is_beam
      damage *= 0.67 --beam attenuation fudge factor
    end

    if w.is_fighter
      damage *= 0.17 --PD and flight time fudge factor
    elseif w.is_launcher
      damage *= 0.33 --PD fudge factor
    end

    score += damage * alive
  end

  return score
end


local valuable_system_types = {
  shield_regenerator = true,
  shield = true,
  capacitor = true,
  reactor = true,
  shuttle = true,
  boarding_module = true
}

local function system_value(s)
  if valuable_system_types[s.info.type]
    return 1
  end
  if s.parent_type == 'Transporter' --Transporter's system type is 'system' for some reason, so it doesn't go in valuable_system_types
    return 1
  end
  if MUNITION[s.parent_type]
    return 0.5 --killing this has a 50% chance to destroy the ship, so return 0.5
  end
  return 0
end

-- a score that represents how "tanky" a potential target is
-- we want this to reflect how much "useless" damage we have to chew through before we start damaging systems
-- so we only include shields and armor
function calc_defence_value(ship)
  -- ignore shield and armour reflection, for simplicity

  local shield_score = 0
  if ship.adjusted_max_shields > 0
    local shield_regen = floor(ship.adjusted_max_shields*ship.shield_regen_rate)*2

    -- use rms to weight highly damaged shield segments more heavily
    -- that way if a shield segment is down it counts more than a straight average
    local rms_shield_dam = 0
    for _,s in pairs(SHIELD_NAMES)
      rms_shield_dam += ship[s]*ship[s]
    end
    rms_shield_dam = sqrt(rms_shield_dam / #SHIELD_NAMES)

    shield_score = ship.max_shields - rms_shield_dam + shield_regen
  end

  local armor_score = ship.armor - ship.armor_damage

  local useless_systems = 0
  for s in systems(ship) do
    if s.damage < s.health -- destroyed systems can't soak up damage
      useless_systems += (1-system_value(s))
    end
  end

  --is there a constant somewhere for the min structure before a ship is destroyed?
  local remaining_structure = max(ship.structure*0.7 - ship.structure_damage, 0)

  return shield_score + armor_score + min(useless_systems, remaining_structure)
end

-- desired engagement ranges
local function desired_weapon_range(w, tracker_survival)
    tracker_survival = tracker_survival or 1

    if w.is_anti_missile
      return ranges.anti_missile
    end
    if w.type=='energy_torpedo'
      return w:launcher_range() * 0.8
    end
    if w.is_fighter
      --want to be within launch range to make an attack every turn, and in case of heavy PD
      --we want to be within roundtrip range (1/2 launch range) to be back in the safety of the
      --hangars before the opponent's turn
      return w:launcher_range()*(0.45 + 0.35*tracker_survival) -- 0.45 to 0.8
    end
    if w.is_launcher
      --In the case of heavy PD, prefer to be within launch range, otherwise keep a fair distance
      return w:launcher_range()*(0.8 + 4.2*tracker_survival) -- 0.8 to 5
    end
    if w.is_pd
      return ranges.short
    end
    if w.is_beam
      --passing 1 to effective_distance() makes it return a range divisor
      return ranges.short/w:effective_distance(1)
    end
    return ranges.medium/w:effective_distance(1)
end

function desired_attack_dist(ship, tracker_survival)
  if ship.weapons?[1]?.parent_type=='Suicide Attack'
    return 3
  end

  local totalweight = 0
  local total = 0

  local totalweight_pd = 0
  local total_pd = 0

  for w in weapons(ship) do
    local range = desired_weapon_range(w, tracker_survival)

    local damage = average_weapon_damage(w) * alive_weapons(w)
    local weight = damage * damage

    if(w.is_pd)
      total_pd += weight * range
      totalweight_pd += weight
    else
      total += weight * range
      totalweight += weight
    end
  end

  --weighted average of all non-pd weapons, unless this ship only has PD weapons
  if(totalweight == 0)
    total = total_pd
    totalweight = totalweight_pd
  end

  --return nil if this ship doesn't attack
  return (totalweight > 0 and total / totalweight) or nil
end
